package ch.epfl.biop.hdf5export;

/**
 * https://github.com/bigdataviewer/bigdataviewer_fiji/blob/master/src/main/java/bdv/ij/ExportImagePlusPlugIn.java
 */

import java.awt.AWTEvent;
import java.awt.Checkbox;
import java.awt.Choice;
import java.awt.TextField;
import java.awt.event.ItemEvent;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Map;

import ij.plugin.ChannelSplitter;
import ij.process.LUT;
import net.imglib2.FinalDimensions;
import net.imglib2.RandomAccessibleInterval;
import net.imglib2.realtransform.AffineTransform3D;

import net.imglib2.util.Intervals;
import org.scijava.command.Command;
import org.scijava.plugin.Plugin;

import bdv.export.ExportMipmapInfo;
import bdv.export.ExportScalePyramid.AfterEachPlane;
import bdv.export.ExportScalePyramid.LoopbackHeuristic;
import bdv.export.ProgressWriter;
import bdv.export.ProposeMipmaps;
import bdv.export.SubTaskProgressWriter;
import bdv.export.WriteSequenceToHdf5;
import bdv.ij.export.imgloader.ImagePlusImgLoader;
import bdv.ij.export.imgloader.ImagePlusImgLoader.MinMaxOption;
import bdv.ij.util.PluginHelper;
import bdv.ij.util.ProgressWriterIJ;
import bdv.img.hdf5.Hdf5ImageLoader;
import bdv.img.hdf5.Partition;
import bdv.spimdata.SequenceDescriptionMinimal;
import bdv.spimdata.SpimDataMinimal;
import bdv.spimdata.XmlIoSpimDataMinimal;
import fiji.util.gui.GenericDialogPlus;
import ij.IJ;
import ij.ImageJ;
import ij.ImagePlus;
import ij.WindowManager;
import ij.gui.DialogListener;
import ij.gui.GenericDialog;
import mpicbg.spim.data.generic.sequence.BasicViewSetup;
import mpicbg.spim.data.registration.ViewRegistration;
import mpicbg.spim.data.registration.ViewRegistrations;
import mpicbg.spim.data.sequence.Channel;
import mpicbg.spim.data.sequence.FinalVoxelDimensions;
import mpicbg.spim.data.sequence.TimePoint;
import mpicbg.spim.data.sequence.TimePoints;
import spimdata.util.Displaysettings;

/**
 * ImageJ plugin to export the current image to xml/hdf5.
 *
 * BIOP : the reason why we duplicated this class is because
 * - we are putting some position metadata in it (see offsets)
 * - we are storing some Brightness / Contrast / Color information in it
 *  - For this, to avoid depending on Bdv Playground, we copied the DisplaySettings class.
 *
 * @author Tobias Pietzsch &lt;tobias.pietzsch@gmail.com&gt;
 */
@Plugin(type = Command.class,
        menuPath = "Plugins>BigDataViewer>Export Current Image as XML/HDF5 (Devils)")
public class ExportDevilsHdf5 implements Command
{
    public static void main( final String[] args )
    {
        new ImageJ();
        IJ.run("Confocal Series (2.2MB)");
        new ExportDevilsHdf5().run();
    }



    @Override
    public void run()
    {
        if ( ij.Prefs.setIJMenuBar )
            System.setProperty( "apple.laf.useScreenMenuBar", "true" );

        // get the current image
        final ImagePlus imp = WindowManager.getCurrentImage();

        // make sure there is one
        if ( imp == null )
        {
            IJ.showMessage( "Please open an image first." );
            return;
        }

        // check the image type
        switch ( imp.getType() )
        {
            case ImagePlus.GRAY8:
            case ImagePlus.GRAY16:
            case ImagePlus.GRAY32:
                break;
            default:
                IJ.showMessage( "Only 8, 16, 32-bit images are supported currently!" );
                return;
        }

        // get calibration and image size
        final double pw = imp.getCalibration().pixelWidth;
        final double ph = imp.getCalibration().pixelHeight;
        final double pd = imp.getCalibration().pixelDepth;

        /**
         * Contains the XYZ position of the dataset in the  source AffineTransform3D of the xml/hdf5 dataset, in pixels
         * Unsupported :
         * - multiangle datasets (like multiangle CZI)
         * Ambiguous :
         * - whether the positions defines the central position of the tile or the top right corner of the tile is not precised using the Bio-Formats API
         * but here, the coordinate given should be the top left corner.
         */
        final double xOffset = imp.getCalibration().xOrigin*pw;
        final double yOffset = imp.getCalibration().yOrigin*ph;
        final double zOffset = imp.getCalibration().zOrigin*pd;

        String punit = imp.getCalibration().getUnit();
        if ( punit == null || punit.isEmpty() )
            punit = "px";
        final FinalVoxelDimensions voxelSize = new FinalVoxelDimensions( punit, pw, ph, pd );
        final int w = imp.getWidth();
        final int h = imp.getHeight();
        final int d = imp.getNSlices();
        final FinalDimensions size = new FinalDimensions( w, h, d );

        // propose reasonable mipmap settings
        final ExportMipmapInfo autoMipmapSettings = ProposeMipmaps.proposeMipmaps( new BasicViewSetup( 0, "", size, voxelSize ) );

        // show dialog to get output paths, resolutions, subdivisions, min-max option
        final Parameters params = getParameters( imp.getDisplayRangeMin(), imp.getDisplayRangeMax(), autoMipmapSettings );
        if ( params == null )
            return;

        final ProgressWriter progressWriter = new ProgressWriterIJ();
        progressWriter.out().println( "starting export..." );

        // create ImgLoader wrapping the image
        final ImagePlusImgLoader< ? > imgLoader;
        switch ( imp.getType() )
        {
            case ImagePlus.GRAY8:
                imgLoader = ImagePlusImgLoader.createGray8( imp, params.minMaxOption, params.rangeMin, params.rangeMax );
                break;
            case ImagePlus.GRAY16:
                imgLoader = ImagePlusImgLoader.createGray16( imp, params.minMaxOption, params.rangeMin, params.rangeMax );
                break;
            case ImagePlus.GRAY32:
            default:
                imgLoader = ImagePlusImgLoader.createGray32( imp, params.minMaxOption, params.rangeMin, params.rangeMax );
                break;
        }

        final int numTimepoints = imp.getNFrames();
        final int numSetups = imp.getNChannels();

        // create SourceTransform from the images calibration
        final AffineTransform3D sourceTransform = new AffineTransform3D();
        sourceTransform.set( pw, 0, 0, xOffset, 0, ph, 0, yOffset, 0, 0, pd, zOffset );

        // write hdf5
        final HashMap< Integer, BasicViewSetup > setups = new HashMap<>( numSetups );

        ImagePlus[] impSingleChannel = ChannelSplitter.split(imp);

        for ( int s = 0; s < numSetups; ++s )
        {
            final BasicViewSetup setup = new BasicViewSetup( s, String.format( "channel %d", s + 1 ), size, voxelSize );
            setup.setAttribute( new Channel( s + 1 ) );

            Displaysettings ds = new Displaysettings(s+1);
            ds.min = impSingleChannel[s].getDisplayRangeMin();
            ds.max = impSingleChannel[s].getDisplayRangeMax();
            if (imp.getType() == ImagePlus.COLOR_RGB) {
                ds.isSet = false;
            } else {
                ds.isSet = true;
                LUT lut = impSingleChannel[s].getProcessor().getLut();
                ds.color = new int[]{lut.getRed(255), lut.getGreen(255), lut.getBlue(255), lut.getAlpha(255)};
            }
            setup.setAttribute(ds);

            setups.put( s, setup );
        }

        impSingleChannel = null;

        final ArrayList< TimePoint > timepoints = new ArrayList<>( numTimepoints );
        for ( int t = 0; t < numTimepoints; ++t )
            timepoints.add( new TimePoint( t ) );
        final SequenceDescriptionMinimal seq = new SequenceDescriptionMinimal( new TimePoints( timepoints ), setups, imgLoader, null );

        Map< Integer, ExportMipmapInfo > perSetupExportMipmapInfo;
        perSetupExportMipmapInfo = new HashMap<>();
        final ExportMipmapInfo mipmapInfo = new ExportMipmapInfo( params.resolutions, params.subdivisions );
        for ( final BasicViewSetup setup : seq.getViewSetupsOrdered() )
            perSetupExportMipmapInfo.put( setup.getId(), mipmapInfo );


        // LoopBackHeuristic:
        // - If saving more than 8x on pixel reads use the loopback image over
        //   original image
        // - For virtual stacks also consider the cache size that would be
        //   required for all original planes contributing to a "plane of
        //   blocks" at the current level. If this is more than 1/4 of
        //   available memory, use the loopback image.
        final boolean isVirtual = imp.getStack().isVirtual();
        final long planeSizeInBytes = imp.getWidth() * imp.getHeight() * imp.getBytesPerPixel();
        final long ijMaxMemory = IJ.maxMemory();
        final int numCellCreatorThreads = Math.max( 1, PluginHelper.numThreads() - 1 );
        final LoopbackHeuristic loopbackHeuristic = new LoopbackHeuristic()
        {
            @Override
            public boolean decide( final RandomAccessibleInterval< ? > originalImg, final int[] factorsToOriginalImg, final int previousLevel, final int[] factorsToPreviousLevel, final int[] chunkSize )
            {
                if ( previousLevel < 0 )
                    return false;

                if ( Intervals.numElements( factorsToOriginalImg ) / Intervals.numElements( factorsToPreviousLevel ) >= 8 )
                    return true;

                if ( isVirtual )
                {
                    final long requiredCacheSize = planeSizeInBytes * factorsToOriginalImg[ 2 ] * chunkSize[ 2 ];
                    if ( requiredCacheSize > ijMaxMemory / 4 )
                        return true;
                }

                return false;
            }
        };

        final AfterEachPlane afterEachPlane = new AfterEachPlane()
        {
            @Override
            public void afterEachPlane( final boolean usedLoopBack )
            {
                if ( !usedLoopBack && isVirtual )
                {
                    final long free = Runtime.getRuntime().freeMemory();
                    final long total = Runtime.getRuntime().totalMemory();
                    final long max = Runtime.getRuntime().maxMemory();
                    final long actuallyFree = max - total + free;

                    if ( actuallyFree < max / 2 )
                        imgLoader.clearCache();
                }
            }

        };

        final ArrayList< Partition > partitions;
        if ( params.split )
        {
            final String xmlFilename = params.seqFile.getAbsolutePath();
            final String basename = xmlFilename.endsWith( ".xml" ) ? xmlFilename.substring( 0, xmlFilename.length() - 4 ) : xmlFilename;
            partitions = Partition.split( timepoints, seq.getViewSetupsOrdered(), params.timepointsPerPartition, params.setupsPerPartition, basename );

            for ( int i = 0; i < partitions.size(); ++i )
            {
                final Partition partition = partitions.get( i );
                final ProgressWriter p = new SubTaskProgressWriter( progressWriter, 0, 0.95 * i / partitions.size() );
                WriteSequenceToHdf5.writeHdf5PartitionFile( seq, perSetupExportMipmapInfo, params.deflate, partition, loopbackHeuristic, afterEachPlane, numCellCreatorThreads, p );
            }
            WriteSequenceToHdf5.writeHdf5PartitionLinkFile( seq, perSetupExportMipmapInfo, partitions, params.hdf5File );
        }
        else
        {
            partitions = null;
            WriteSequenceToHdf5.writeHdf5File( seq, perSetupExportMipmapInfo, params.deflate, params.hdf5File, loopbackHeuristic, afterEachPlane, numCellCreatorThreads, new SubTaskProgressWriter( progressWriter, 0, 0.95 ) );
        }

        // write xml sequence description
        final Hdf5ImageLoader hdf5Loader = new Hdf5ImageLoader( params.hdf5File, partitions, null, false );
        final SequenceDescriptionMinimal seqh5 = new SequenceDescriptionMinimal( seq, hdf5Loader );

        final ArrayList< ViewRegistration > registrations = new ArrayList<>();
        for ( int t = 0; t < numTimepoints; ++t )
            for ( int s = 0; s < numSetups; ++s )
                registrations.add( new ViewRegistration( t, s, sourceTransform ) );

        final File basePath = params.seqFile.getParentFile();
        final SpimDataMinimal spimData = new SpimDataMinimal( basePath, seqh5, new ViewRegistrations( registrations ) );

        try
        {
            new XmlIoSpimDataMinimal().save( spimData, params.seqFile.getAbsolutePath() );
            progressWriter.setProgress( 1.0 );
        }
        catch ( final Exception e )
        {
            throw new RuntimeException( e );
        }
        progressWriter.out().println( "done" );
    }

    protected static class Parameters
    {
        final boolean setMipmapManual;

        final int[][] resolutions;

        final int[][] subdivisions;

        final File seqFile;

        final File hdf5File;

        final MinMaxOption minMaxOption;

        final double rangeMin;

        final double rangeMax;

        final boolean deflate;

        final boolean split;

        final int timepointsPerPartition;

        final int setupsPerPartition;

        public Parameters(
                final boolean setMipmapManual, final int[][] resolutions, final int[][] subdivisions,
                final File seqFile, final File hdf5File,
                final MinMaxOption minMaxOption, final double rangeMin, final double rangeMax, final boolean deflate,
                final boolean split, final int timepointsPerPartition, final int setupsPerPartition )
        {
            this.setMipmapManual = setMipmapManual;
            this.resolutions = resolutions;
            this.subdivisions = subdivisions;
            this.seqFile = seqFile;
            this.hdf5File = hdf5File;
            this.minMaxOption = minMaxOption;
            this.rangeMin = rangeMin;
            this.rangeMax = rangeMax;
            this.deflate = deflate;
            this.split = split;
            this.timepointsPerPartition = timepointsPerPartition;
            this.setupsPerPartition = setupsPerPartition;
        }
    }

    static boolean lastSetMipmapManual = false;

    static String lastSubsampling = "{1,1,1}, {2,2,1}, {4,4,2}";

    static String lastChunkSizes = "{32,32,4}, {16,16,8}, {8,8,8}";

    static int lastMinMaxChoice = 2;

    static double lastMin = 0;

    static double lastMax = 65535;

    static boolean lastSplit = false;

    static int lastTimepointsPerPartition = 0;

    static int lastSetupsPerPartition = 0;

    static boolean lastDeflate = true;

    static String lastExportPath = "./export.xml";

    protected Parameters getParameters( final double impMin, final double impMax, final ExportMipmapInfo autoMipmapSettings  )
    {
        if ( lastMinMaxChoice == 0 ) // use ImageJs...
        {
            lastMin = impMin;
            lastMax = impMax;
        }

        while ( true )
        {
            final GenericDialogPlus gd = new GenericDialogPlus( "Export for BigDataViewer" );

            gd.addCheckbox( "manual_mipmap_setup", lastSetMipmapManual );
            final Checkbox cManualMipmap = ( Checkbox ) gd.getCheckboxes().lastElement();
            gd.addStringField( "Subsampling_factors", lastSubsampling, 25 );
            final TextField tfSubsampling = ( TextField ) gd.getStringFields().lastElement();
            gd.addStringField( "Hdf5_chunk_sizes", lastChunkSizes, 25 );
            final TextField tfChunkSizes = ( TextField ) gd.getStringFields().lastElement();

            gd.addMessage( "" );
            final String[] minMaxChoices = new String[] { "Use ImageJ's current min/max setting", "Compute min/max of the (hyper-)stack", "Use values specified below" };
            gd.addChoice( "Value_range", minMaxChoices, minMaxChoices[ lastMinMaxChoice ] );
            final Choice cMinMaxChoices = (Choice) gd.getChoices().lastElement();
            gd.addNumericField( "Min", lastMin, 0 );
            final TextField tfMin = (TextField) gd.getNumericFields().lastElement();
            gd.addNumericField( "Max", lastMax, 0 );
            final TextField tfMax = (TextField) gd.getNumericFields().lastElement();

            gd.addMessage( "" );
            gd.addCheckbox( "split_hdf5", lastSplit );
            final Checkbox cSplit = ( Checkbox ) gd.getCheckboxes().lastElement();
            gd.addNumericField( "timepoints_per_partition", lastTimepointsPerPartition, 0, 25, "" );
            final TextField tfSplitTimepoints = ( TextField ) gd.getNumericFields().lastElement();
            gd.addNumericField( "setups_per_partition", lastSetupsPerPartition, 0, 25, "" );
            final TextField tfSplitSetups = ( TextField ) gd.getNumericFields().lastElement();

            gd.addMessage( "" );
            gd.addCheckbox( "use_deflate_compression", lastDeflate );

            gd.addMessage( "" );
            PluginHelper.addSaveAsFileField( gd, "Export_path", lastExportPath, 25 );

//			gd.addMessage( "" );
//			gd.addMessage( "This Plugin is developed by Tobias Pietzsch (pietzsch@mpi-cbg.de)\n" );
//			Bead_Registration.addHyperLinkListener( ( MultiLineLabel ) gd.getMessage(), "mailto:pietzsch@mpi-cbg.de" );

            final String autoSubsampling = ProposeMipmaps.getArrayString( autoMipmapSettings.getExportResolutions() );
            final String autoChunkSizes = ProposeMipmaps.getArrayString( autoMipmapSettings.getSubdivisions() );
            gd.addDialogListener( new DialogListener()
            {
                @Override
                public boolean dialogItemChanged( final GenericDialog dialog, final AWTEvent e )
                {
                    gd.getNextBoolean();
                    gd.getNextString();
                    gd.getNextString();
                    gd.getNextChoiceIndex();
                    gd.getNextNumber();
                    gd.getNextNumber();
                    gd.getNextBoolean();
                    gd.getNextNumber();
                    gd.getNextNumber();
                    gd.getNextBoolean();
                    gd.getNextString();
                    if ( e instanceof ItemEvent && e.getID() == ItemEvent.ITEM_STATE_CHANGED && e.getSource() == cMinMaxChoices )
                    {
                        final boolean enable = cMinMaxChoices.getSelectedIndex() == 2;
                        tfMin.setEnabled( enable );
                        tfMax.setEnabled( enable );
                    }
                    else if ( e instanceof ItemEvent && e.getID() == ItemEvent.ITEM_STATE_CHANGED && e.getSource() == cManualMipmap )
                    {
                        final boolean useManual = cManualMipmap.getState();
                        tfSubsampling.setEnabled( useManual );
                        tfChunkSizes.setEnabled( useManual );
                        if ( !useManual )
                        {
                            tfSubsampling.setText( autoSubsampling );
                            tfChunkSizes.setText( autoChunkSizes );
                        }
                    }
                    else if ( e instanceof ItemEvent && e.getID() == ItemEvent.ITEM_STATE_CHANGED && e.getSource() == cSplit )
                    {
                        final boolean split = cSplit.getState();
                        tfSplitTimepoints.setEnabled( split );
                        tfSplitSetups.setEnabled( split );
                    }
                    return true;
                }
            } );

            final boolean enable = lastMinMaxChoice == 2;
            tfMin.setEnabled( enable );
            tfMax.setEnabled( enable );

            tfSubsampling.setEnabled( lastSetMipmapManual );
            tfChunkSizes.setEnabled( lastSetMipmapManual );
            if ( !lastSetMipmapManual )
            {
                tfSubsampling.setText( autoSubsampling );
                tfChunkSizes.setText( autoChunkSizes );
            }

            tfSplitTimepoints.setEnabled( lastSplit );
            tfSplitSetups.setEnabled( lastSplit );

            gd.showDialog();
            if ( gd.wasCanceled() )
                return null;

            lastSetMipmapManual = gd.getNextBoolean();
            lastSubsampling = gd.getNextString();
            lastChunkSizes = gd.getNextString();
            lastMinMaxChoice = gd.getNextChoiceIndex();
            lastMin = gd.getNextNumber();
            lastMax = gd.getNextNumber();
            lastSplit = gd.getNextBoolean();
            lastTimepointsPerPartition = ( int ) gd.getNextNumber();
            lastSetupsPerPartition = ( int ) gd.getNextNumber();
            lastDeflate = gd.getNextBoolean();
            lastExportPath = gd.getNextString();

            // parse mipmap resolutions and cell sizes
            final int[][] resolutions = PluginHelper.parseResolutionsString( lastSubsampling );
            final int[][] subdivisions = PluginHelper.parseResolutionsString( lastChunkSizes );
            if ( resolutions.length == 0 )
            {
                IJ.showMessage( "Cannot parse subsampling factors " + lastSubsampling );
                continue;
            }
            if ( subdivisions.length == 0 )
            {
                IJ.showMessage( "Cannot parse hdf5 chunk sizes " + lastChunkSizes );
                continue;
            }
            else if ( resolutions.length != subdivisions.length )
            {
                IJ.showMessage( "subsampling factors and hdf5 chunk sizes must have the same number of elements" );
                continue;
            }

            final MinMaxOption minMaxOption;
            if ( lastMinMaxChoice == 0 )
                minMaxOption = MinMaxOption.TAKE_FROM_IMAGEPROCESSOR;
            else if ( lastMinMaxChoice == 1 )
                minMaxOption = MinMaxOption.COMPUTE;
            else
                minMaxOption = MinMaxOption.SET;

            String seqFilename = lastExportPath;
            if ( !seqFilename.endsWith( ".xml" ) )
                seqFilename += ".xml";
            final File seqFile = new File( seqFilename );
            final File parent = seqFile.getParentFile();
            if ( parent == null || !parent.exists() || !parent.isDirectory() )
            {
                IJ.showMessage( "Invalid export filename " + seqFilename );
                continue;
            }
            final String hdf5Filename = seqFilename.substring( 0, seqFilename.length() - 4 ) + ".h5";
            final File hdf5File = new File( hdf5Filename );

            return new Parameters( lastSetMipmapManual, resolutions, subdivisions, seqFile, hdf5File, minMaxOption, lastMin, lastMax, lastDeflate, lastSplit, lastTimepointsPerPartition, lastSetupsPerPartition );		}
    }
}

